/* =====================================================================
 * NetArduino for showing weather forecast and time from the internet
 * ---------------------------------------------------------------------
 * NTPAndTime.cpp
 *   Call and maintain time through NTP, show on console
 *   and send over serial to the serial display
 * =====================================================================
 * By: Zakie Mashiah
 * =====================================================================
 * Copyright
 *  You may use this code with single condition of giving visible
 *  credit to the author.
 * =====================================================================
 * TODO
 * =====================================================================
 *  Revision History
 *      31-Oct-2011 v0.9
 * =====================================================================
 */
#include <avr/pgmspace.h>
#include <SPI.h>         
#include <Ethernet.h>
#include <Udp.h>
#include <EthernetDNS.h>

#include "CTimeLib.h"

#include "utilfuncs.h"
#include "NTPAndTime.h"

#define DST_PUSHBUTTON_PIN 2

static const int 		NTP_PACKET_SIZE = 48; // NTP time stamp is in the first 48 bytes of the message
static byte 	 		packetBuffer[NTP_PACKET_SIZE+16]; //buffer to hold incoming and outgoing packets 
static unsigned int		localPort = 8888;      // local port to listen for UDP packets
static char 			*timeServerName = "time.nist.gov";
static byte 		 	timeServerIP[4];
static unsigned long timeZoneSeconds = 2 * 60 * 60; // Set according to your timezone. Negative value if you are GMT-x
static const unsigned long dsSeconds = 60 * 60;  // How much is your DST?
static volatile byte       daylightSavings = 0; // 1 if in DST, 0 if not
static volatile byte       dstChanged = 0;		// 1 if button was pushed and daylightSavings was not sent to display


// Called by interrupt of button pushed
void ToggleDST()
{
	daylightSavings = !daylightSavings; // Toggle the state of DST
	dstChanged = 1;
}


// Checks if need to show annd send time again, if so it does
void TimeUpdateRequired()
{
	if (dstChanged)
	{
		dstChanged = 0;
		setForceSync();
		SendTime();
		ShowTime();
	}
}


void PrintLeading0(int n)
{
	if (n < 10)
		Console.print('0');
	Console.print(n);
}

// send an NTP request to the time server at the given address 
static unsigned long sendNTPpacket(byte *address)
{
	// set all bytes in the buffer to 0
	memset(packetBuffer, 0, sizeof(packetBuffer));
	// Initialize values needed to form NTP request
	// (see URL above for details on the packets)
	packetBuffer[0] = 0b11100011;   // LI, Version, Mode
	packetBuffer[1] = 0;     		// Stratum, or type of clock
	packetBuffer[2] = 6;     		// Polling Interval
	packetBuffer[3] = 0xEC;  		// Peer Clock Precision
	// 8 bytes of zero for Root Delay & Root Dispersion
	packetBuffer[12]  = 49;
	packetBuffer[13]  = 0x4E;
	packetBuffer[14]  = 49;
	packetBuffer[15]  = 52;

	// all NTP fields have been given values, now
	// you can send a packet requesting a timestamp:
	Udp.sendPacket(packetBuffer, NTP_PACKET_SIZE, address, 123); //NTP requests are to port 123
}

static unsigned long word2long(unsigned long highWord, unsigned long lowWord)
{
	unsigned long ul;
	
	ul = highWord;
	ul <<= 16;
	ul |= lowWord;
	return ul;
}

time_t getNtpTime()
{
  int bytesRead;
  
	sendNTPpacket(timeServerIP); // send an NTP packet to a time server
	Console_printP(PSTR("NTP request sent"));
	// wait to see if a reply is available
	if ( Udp.available() )
	{
		bytesRead = Udp.readPacket(packetBuffer, NTP_PACKET_SIZE);  // read the packet into the buffer
		if (bytesRead != NTP_PACKET_SIZE)
        {
            Console_printP(PSTR(" Did not recieve enough bytes"));
            return 0;
        }
		Console_printlnP(PSTR(" Got NTP reply"));
		
		//the timestamp starts at byte 40 of the received packet and is four bytes,
		// or two words, long. First, esxtract the two words:
		unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
		unsigned long lowWord  = word(packetBuffer[42], packetBuffer[43]);
		
		// combine the four bytes (two words) into a long integer
		// this is NTP time (seconds since Jan 1 1900):
		unsigned long secsSince1900 = word2long(highWord, lowWord);
	
		// now convert NTP time into everyday time:
		// Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
		const unsigned long seventyYears = 2208988800UL;
		
		// subtract seventy years:
		// Add TimeZone data, Tel-Aviv is two hours before GMT
		unsigned long tzOffset = timeZoneSeconds + (daylightSavings * dsSeconds);
        Console_printP(PSTR("TZ=")); Console.println(tzOffset);
		unsigned long epoch = secsSince1900 - seventyYears + tzOffset;

		Console_printP(PSTR("Unix time = "));
		Console.println(epoch);
		
		return epoch;
	}
	return 0;
}


void InitializeNTP()
{
	// Get the IP address of timeServerName
	DNSError netError = EthernetDNS.resolveHostName(timeServerName, timeServerIP);
	if (netError == DNSSuccess)
	{
		Console_printP(PSTR("Got NTP IP address of "));
		Console.print(timeServerName);
		Console.print(" ");
	}
	else
	{
		timeServerIP[0] = 192;
		timeServerIP[1] =  43;
		timeServerIP[2] = 244;
		timeServerIP[3] =  18;
		Console_printP(PSTR("Using default IP address for NTP server"));
	}
    Console.println(ip_to_str(timeServerIP));
	
	// start  UDP socket
	Udp.begin(localPort);

	// Sets time
	setSyncProvider(getNtpTime);
	//setSyncInterval(60);
	Console_printlnP(PSTR("Set SyncTimeProvider"));
	
	for (int count=1; count<30; count++)
	{
        if ((count % 10) == 0)
            Console_printlnP(PSTR("... still waiting for NTP"));
		now();
		if ( timeStatus()== timeNotSet)
			continue;
		else
			break;
	}
	// DST pushbutton settings
	pinMode(DST_PUSHBUTTON_PIN, INPUT);
	attachInterrupt(0, ToggleDST, FALLING);
}




void ShowTime()
{
	static unsigned long currentTime;
	static const char column = ':';
	static const char slash  = '/';
	
	if ( now() != currentTime)
	{
		currentTime = now();
		Console.print(hour());
		Console.print(column);
		PrintLeading0(minute());
		Console.print(column);
		PrintLeading0(second());
		Console.print(" ");
		Console.print(day());
		Console.print(slash);
		Console.print(month());
		Console.print(slash);
		Console.print(year()); 
		Console.println(); 
	}
}


void SendTime()
{
	char buff[30];
	static time_t last;
	time_t tNow;

	do
	{
		tNow = now();
		if (tNow < last)
			if ( (last - tNow) < 600) // Less than 10 minutes is an error
			{
				time_t t;
				
				Console_printlnP(PSTR("GOT BACK IN TIME, retry..."));
				t = getNtpTime();
				if (t != 0UL)
					setTime(t);
				else
					delay(1000);
			}
			else
				break;
	} while (tNow < last);
	last = tNow;

	if (timeStatus() == timeSet)
	{
		sprintf(buff, "[T%4d/%d/%d %d:%d:%d]", year(), month(), day(), hour(), minute(), second());
		serialDisplay.print(buff);
	}
}