/* =====================================================================
 * NetArduino for showing weather forecast and time from the internet
 * ---------------------------------------------------------------------
 * Program is using Ethernet Arduino to access the internet and reads
 * weather data and forecast from Google weather. Program also access
 * NTP servers over the internet to get current time.
 * 
 * Date is send over UART port to a serial display but also shown on
 * Arduino serial output (we call it Console here) for debugging.
 *
 * By deault the program uses DHCP to initialize itself in the network,
 * but you can disable that by simple change of a #define (see below).
 *
 * Due to very tight memory, the program needs to use an ATMega trick
 * of storing some data in program space. This makes the code a bit
 * hard to read with PSTR macro being used pretty often. Somehow it
 * feels like a miracle to read an XML file from Ethernet and be able
 * to process it in 2KB RAM the Arduino has. My first attempt at doing
 * very similar device with Netduino (Fez Panda II) resulted in
 * out-of-memory in much earlier stages. Without using program space
 * strings I got very strange behaviour of resets due to overflow of
 * data. So as much as I dislike the way the code needs to be, there
 * was no other choice.
 *
 *
 * 
 * =====================================================================
 * By: Zakie Mashiah
 * =====================================================================
 * Copyright
 *  You may use this code with single condition of giving visible
 *  credit to the author.
 * =====================================================================
 * TODO
 * =====================================================================
 *  Revision History
 *      31-Oct-2011 v0.91
 *		09-Nov-2011 v0.95 Add the indoor DHT22 sensor support
 * =====================================================================
 */
#include <avr/pgmspace.h>
#include <SPI.h>         
#include <Ethernet.h>
#include <EthernetDHCP.h>
#include <EthernetDNS.h>
#include <Udp.h>
#include <Client.h>
#include <NewSoftSerial.h>
#include <dht.h>

#include "utilfuncs.h"
#include "NTPAndTime.h"
#include "weatherdata.h"
#include "dhtsensor.h"

#define USE_DHCP      1		// Set to 0 is you want static IPs
#define USE_NTP       1		// Set to 0 if you do not want NTP (why?)
#define SHOW_NET_INFO 1		// Set to 0 to reduce debug messages of network setup


NewSoftSerial secondSerial(4,5);

// Enter a MAC address and IP address for your controller below.
byte mac[] = {   0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xE0 };

#define MAX_STRING_LEN (256*2*1)
char dataStr[MAX_STRING_LEN];
int  xmlStreamByteCount;

// Flags to differentiate XML tags from document elements (ie. data)
enum TagSection { XmlStart = 0, XmlForecastInformation, XmlCurrentConditions, XmlForecastConditions } currentSection = XmlStart;


static const char weatherServerName[]	 = "www.google.com";
static const char weatherDataGetString[] = "GET /ig/api?weather=Tel-Aviv HTTP/1.0";
static byte weatherServerIP[4];

Client weatherClient(0, 80);
static dhtSensor indoorSensor;

#if USE_DHCP
	// Nothing to do here, DHCP will tell us who we are and where to go to
#else
const byte ipAddr[]      = { 192, 168, 2, 7};;  // At my home 192.168.2.[2-99] are not allocated by DHCP on DSL router
const byte gatewayAddr[] = { 192, 168, 2, 1};	// At my home this is the DSL router IP
const byte dnsAddr[]     = { 192, 168, 2, 1};	// And of course it is also the DNS
#endif

WeatherData weatherData;		// Stores the weather and forecast information from Google Weather service
bool setupHadErrors = false;	// if true, main loop will reset

void ConcatChar(char *str, char ch);
int matchXmlTag (char *str, char* searchTag);


static void initNetwork()
{
	Console_printlnP(PSTR("-> initNetwork"));

#if USE_DHCP
	// Initiate a DHCP session. The argument is the MAC (hardware) address that
	// you want your Ethernet shield to use. This call will block until a DHCP
	// lease has been obtained. The request will be periodically resent until
	// a lease is granted, but if there is no DHCP server on the network or if
	// the server fails to respond, this call will block forever.
	// Thus, you can alternatively use polling mode to check whether a DHCP
	// lease has been obtained, so that you can react if the server does not
	// respond (see the PollingDHCP example).
	EthernetDHCP.begin(mac);
	Console_printlnP(PSTR("--> EthernetDHCP"));

	// Since we're here, it means that we now have a DHCP lease, so we print
	// out some information.
	const byte* ipAddr 		= EthernetDHCP.ipAddress();
	const byte* gatewayAddr = EthernetDHCP.gatewayIpAddress();
	const byte* dnsAddr 	= EthernetDHCP.dnsIpAddress();
	Console_printlnP(PSTR("--> A DHCP lease has been obtained."));
#endif
#if SHOW_NET_INFO
	Console_printP(PSTR("--> My IP address = "));
	Console.println(ip_to_str(ipAddr));
	Console_printP(PSTR("--> Gateway IP address = "));
	Console.println(ip_to_str(gatewayAddr));
	Console_printP(PSTR("--> DNS IP address = "));
	Console.println(ip_to_str(dnsAddr));
#endif
#if !(USE_DHCP)
	Ethernet.begin(mac, (uint8_t *)ipAddr);
#endif
	Console_printlnP(PSTR("<-- Network IP addresses complete"));

	Console_printlnP(PSTR("--> Setting DNS"));
	EthernetDNS.setDNSServer(dnsAddr);
	Console_printlnP(PSTR("<-- DNS set"));

	Console_printlnP(PSTR("<- initNetwork complete"));
}

void (*restart)(void) = 0;

void setup() 
{
	Console.begin(9600);
	Console_printlnP(PSTR("Staring program..."));

	serialDisplay.begin(31250);
	initNetwork();
#if USE_NTP
	InitializeNTP();
#endif
	
	// Get the IP address of weather service
	Console_printP(PSTR("--> Resolving hostanme to IP for ["));
	Console.print((weatherServerName));
	Console_printlnP(PSTR("]"));
	DNSError netError = EthernetDNS.resolveHostName((weatherServerName), weatherServerIP);
	if (netError == DNSSuccess)
	{
		Console_printP(PSTR("<--- Got IP address of ["));
		Console.print(weatherServerName);
        Console_printP(PSTR("] "));
	}
	else
	{
		Console_printlnP(PSTR("<--- ERROR: could not get IP address"));
		setupHadErrors = true;
        return;
	}
    Console.println(ip_to_str(weatherServerIP));
    Console_printlnP(PSTR("<--- ready to connect"));
	weatherClient.Init(weatherServerIP, 80);

	_ConnectToGoogleWeather();
	
	indoorSensor.init();	// Initialize the indoor DHT22 sensor
	
	Console_printlnP(PSTR("setup() complete"));
}

void loop()
{
	static char counter = 0;
	
	Console_printlnP(PSTR("In loop"));
#if USE_DHCP
	EthernetDHCP.maintain(); // To maintain the lease from DHCP server
#endif
#if USE_NTP
	ShowTime();		// On console
	SendTime();		// To the serial display
#endif
	if (setupHadErrors == 0)
	{
		WeatherForecast();
        indoorSensor.print();
        indoorSensor.send();
	}
	else
	{
        counter ++;
        if (counter >= 5)
        {
			counter = 0;
			restart();
        }
		else
			// Say we have a prolem
			Console_printlnP(PSTR("ERROR: Please RESET ME"));
	}

	Console_printP(PSTR("Sleeping"));
	for(int i=0; i<10; i++) // 1 minute
	{
		TimeUpdateRequired();
		delay(1000 * 6);
		Console.print(i);	// But show progress every 6 seconds
		TimeUpdateRequired();
	}
	Console.println();
	_ConnectToGoogleWeather();
}

void WeatherForecast()
{
	Console_printlnP(PSTR("---> calling GET from server"));
	weatherClient.println((weatherDataGetString));
	weatherClient.println();
	delay(2000);
	
	// Read serial data in from web:
	xmlStreamByteCount = 0;
	weatherData.Reset();
	while (weatherClient.available())
		ProcessWeatherStream();
	Console.println(xmlStreamByteCount);
	if (xmlStreamByteCount)
	{
		weatherData.Print();
		weatherData.Send();
	}
	weatherClient.stop();
}


void _ConnectToGoogleWeather()
{
	Console_printP(PSTR("---> Connecting to ")); Console.println((weatherServerName));
	if (weatherClient.connect())
	{
		Console_printlnP(PSTR("<--- Connected"));
	}
	else
	{
		Console_printlnP(PSTR("<--- Connection failed"));
		setupHadErrors = true;
	}
}

// Process each char from web
void ProcessWeatherStream()
{
	// Read a char
	int inChar = weatherClient.read();
	xmlStreamByteCount++;
	if (inChar > 0)
    {
		if ( (inChar != 10) && (inChar != '>') )
		{
			ConcatChar(dataStr, inChar);
		}
		else
		{
			_ProcessWeatherXMLLine();
            memset(dataStr, 0, sizeof(dataStr));
		}
    }
}

void _ProcessWeatherXMLLine()
{
	int location;
	char *ptr = dataStr;

	for (ptr = dataStr; *ptr && ((*ptr == ' ') || (*ptr == '-')); ptr++); // Skip leading spaced
	if ( *ptr != '<')
        return;

	//Console.println(ptr);

	switch (currentSection)
	{
		case XmlStart:
			// Search for forecast_information
			if (matchXmlTag(ptr, PSTR("<forecast_information")))
			{
				currentSection = XmlForecastInformation;
			}
			break;
		case XmlForecastInformation:
			if ((location = matchXmlTag(ptr, PSTR("<city data="))) != 0)
			{
				ptr += location;
				weatherData.setCity(ParseCity(ptr));
				currentSection = XmlCurrentConditions;
			}
			break;
		case XmlCurrentConditions:
			if ((location = matchXmlTag(ptr, PSTR("<temp_c data="))) != 0)
			{
				ptr += location;
				weatherData.setTempCurr(ParseTemp(ptr));
			}
			else if ((location = matchXmlTag(ptr, PSTR("<humidity data="))) != 0)
			{
				ptr += location;
				weatherData.setHumidCurr(ParseHumidity(ptr));
			}
			if ((location = matchXmlTag(ptr, PSTR("<icon data="))) != 0)
			{
				ptr += location;
				weatherData.setIconCurr(ParseIcon(ptr));
			}
			else if ((location = matchXmlTag(ptr, PSTR("<wind_condition data="))) != 0)
			{
				ptr += location;
				weatherData.setWindCurr(ParseWind(ptr));
			}
			else if ((location = matchXmlTag(ptr, PSTR("<condition data="))) != 0)
			{
				ptr += location;
				weatherData.setCondCurr(ParseCond(ptr));
			}
			else if ((location = matchXmlTag(ptr, PSTR("</current_conditions"))) != 0)
				currentSection = XmlForecastConditions;
			break;
		case XmlForecastConditions:
			if ((location = matchXmlTag(ptr, PSTR("</xml_api_reply"))) != 0) // end of the stream
			{
				currentSection = XmlStart;
				return;
			}
			else if ((location = matchXmlTag(ptr, PSTR("</forecast_conditions"))) != 0)
			{
				// New member to the array of forecast days
				weatherData.addForecast();
			}
			else if ((location = matchXmlTag(ptr, PSTR("<day_of_week data="))) != 0)
			{
				ptr += location;
				weatherData.setDowFore(ParseDow(ptr));
			}
			else if ((location = matchXmlTag(ptr, PSTR("<low data="))) != 0)
			{
				ptr += location;
				weatherData.setLowTempFore(Far2Cel(ParseTemp(ptr)));
			}
			else if ((location = matchXmlTag(ptr, PSTR("<high data="))) != 0)
			{
				ptr += location;
				weatherData.setHighTempFore(Far2Cel(ParseTemp(ptr)));
			}
			else if ((location = matchXmlTag(ptr, PSTR("<icon data="))) != 0)
			{
				ptr += location;
				weatherData.setIconFore(ParseIcon(ptr));
			}
			else if ((location = matchXmlTag(ptr, PSTR("<condition data="))) != 0)
			{
				ptr += location;
				weatherData.setCondFore(ParseCond(ptr));
			}
			break;
	}
}



/* ----------------------------------------------
 * Other Functions 
 * ---------------------------------------------- */
//Function to add a char to a string and check its length
void ConcatChar(char *str, char ch)
{
	int pos;

	pos = strlen(str);
	
	if (strlen(str) > MAX_STRING_LEN - 2)
	{
		Console_printlnP(PSTR("+++++ ERROR: oversided string"));
		pos = 0;
	}
	str[pos] = ch;
}



// Function to check the current tag for a specific string
int matchXmlTag (char *str, char* searchTag)
{
	int len = strlen(str);
	int pos = strlen_P(searchTag);

	if (len < pos)
		return 0;
	if ( strncmp_P(str, searchTag, pos) == 0 )
		return pos;
	else
		return 0;
}

int isnum(char c)
{
	if (c >= '0')
		if (c <= '9')
			return 1;
	return 0;
}



int Far2Cel(int t)
{
	double d = t;
	d -= 32.0;
	d *= 5.0;
	d/= 9.0;
	
	return d;
}



/* ---------------------------------------------------------
 * Parsing functions for various fields from the XML stream
 * --------------------------------------------------------- */
char *ParseCity(char *s)
{
	char *ptr = s;
	char *city;
	
	ptr++; // Skip the "
	city = ptr;
	for( ; *ptr ; ptr++)
		if (*ptr == '"') // nul the ending "
			*ptr = 0;
	return city;
}



int ParseHumidity(char *s)
{
	char *ptr;
        char *ptr1;

    for(ptr=s; *ptr && (isnum(*ptr) == 0); ptr++); // Skip leading characters
    for( ptr1=ptr; *ptr1 && isnum(*ptr1); ptr1++);
    if ( *ptr1)
		*ptr1 = 0; // Clear the % and rest of chars

    return ParseTemp(ptr);
}

char *ParseIcon(char *s)
{
	char *ptr = s;
	char *last_slash;

	if (*ptr == '"')
        ptr++;
    last_slash = ptr;

	for ( ; *ptr && (*ptr != '"') ; ptr++)
		if (*ptr == '/')
			last_slash = ptr;
	for(ptr = last_slash; *ptr; ptr++)
		if (*ptr == '"')
			*ptr = 0;

	return last_slash + 1;
}



int ParseTemp(char *s)
{
	int temp;
	char *ptr;
	char *ptr1;
	
	for(ptr = s; *ptr && (isnum(*ptr) == 0); ptr++); // skip leading characters;
	for(ptr1 = ptr ; *ptr1 && isnum(*ptr1); ptr1++);
	*ptr1 = 0; // terminate string
	
	for(temp =0; *ptr; ptr++)
	{
		temp *= 10;
		temp += (*ptr) - '0';
	}
	return temp;
}

char *ParseDow(char *s)
{
	char *ptr = s;
	
	if (*ptr == '"')
		ptr++;
        if (ptr[3] == '"')
          ptr[3] = 0;
	return ptr;
}



char *ParseCond(char *s)
{
	char *base;
	char *ptr;
	
	base = s;
	if (base[0] == '"')
		base++;
	for(ptr=base; *ptr; ptr++); // Get to the end of the string
	for(ptr--; ((*ptr == '/') || (*ptr == '"')) && (ptr > base); ptr--);
	ptr++;
	*ptr = 0;

	return base;
}



char *ParseWind(char *s)
{
	char *base = s;
	char *ptr;
	
	if (base[0] == '"')
		base++;
	
	for (ptr = base; *ptr && (*ptr != ' '); ptr++); // look for space
	ptr++;

    base = ptr;
	// Clear trailing " and /
	for(; *ptr; ptr++); // Get to the end of the string
	for(ptr--; ((*ptr == '/') || (*ptr == '"')) && (ptr > base); ptr--);
	ptr++;
	*ptr = 0;
	return base;
}